package MIME::Decoder::XX;


=head1 NAME

MIME::Decoder::XX - decode a "xxencoded" stream


=head1 SYNOPSIS

A generic decoder object; see L<MIME::Decoder> for usage.

Also supports a preamble() method to recover text before
the uuencoded portion of the stream.


=head1 DESCRIPTION

A MIME::Decoder subclass for a nonstandard encoding whereby
data are xxencoded.  Common non-standard MIME encodings for this:

    x-xx
    x-xxencode


=head1 AUTHOR

XX-decoding code adapted from MIME::Decoder::UU (this package).

All rights reserved.  This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

=head1 VERSION

$Revision: 1.11 $ $Date: 2006/03/17 21:03:23 $

=cut


require 5.002;
use vars qw(@ISA $VERSION);
use MIME::Decoder;
use MIME::Tools qw(whine);
use POSIX qw(ceil);
use strict;

@ISA = qw(MIME::Decoder);

# The package version, both in 1.23 style *and* usable by MakeMaker:
$VERSION = "5.420";

my @ENC = split('','+-0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz');

sub DEC_MAP {
    my $charin = shift;
    my $charout;

    if ($charin >= 97) {                     # [a-z]
        $charout = $charin - 59;
    }
    elsif ($charin >= 65) {                  # [A-Z]
        $charout = $charin - 53;
    }
    elsif ($charin >= 48) {                  # [0-9]
        $charout = $charin - 46;
    }
    else {
        $charout = ($charin == 45) ? 1 : 0;  # [\-+]
    }
    return $charout;
}

#------------------------------
#
# decode_it IN OUT
#
sub decode_it {
    my ($self, $in, $out) = @_;
    my ($mode, $file);
    my @preamble;

    ### Init:
    $self->{MDX_Preamble} = \@preamble;
    $self->{MDX_Mode} = undef;
    $self->{MDX_File} = undef;

    ### Find beginning...
    local $_;
    while (defined($_ = $in->getline)) {
	if (/^begin(.*)/) {        ### found it: now decode it...
	    my $modefile = $1;
	    if ($modefile =~ /^(\s+(\d+))?(\s+(.*?\S))?\s*\Z/) {
		($mode, $file) = ($2, $4);
	    }
	    last;                  ### decoded or not, we're done
	}
	push @preamble, $_;
    }
    die("xx decoding: no begin found\n") if !defined($_);      # hit eof!

    ### Store info:
    $self->{MDX_Mode} = $mode;
    $self->{MDX_File} = $file;

    ### Decode:
    while (defined($_ = $in->getline)) {
	last if /^end/;
	next unless /^[0-9a-zA-Z+\-]+$/;
        next unless int(( DEC_MAP(ord()) + 2) / 3) == int(length() / 4);
	$out->print(_decode($_));
    }
    ### chmod oct($mode), $file;    # sheeyeah... right...
    whine "file incomplete, no end found\n" if !defined($_); # eof
    1;
}

#------------------------------
#
# encode_it IN, OUT
#
sub encode_it {
    my ($self, $in, $out) = @_;
    my $buf = '';
    my $lines = 0;

    my $fname = (($self->head && 
		  $self->head->mime_attr('content-disposition.filename')) ||
		 '');
    $out->print("begin 644 $fname\n");
    while ($in->read($buf, 45)) {
        $out->print(_encode($buf));
        $lines ++;
    }
    # If theres data, put a blank line
    $out->print(_encode('')) if $lines;
    $out->print("end\n");
    1;
}

sub _encode {
    my $in = shift;
    my $out = '';

    my $input_count = length($in);
    for (my $offset = 0; $offset < $input_count; $offset += 3) {
       my @chunk = unpack("C3", substr($in, $offset, 3));
       my ($bshift, $val, $i);

       # 3 bytes (8 bits)
       $bshift = 24;
       foreach (@chunk) {
          $bshift -= 8;
          $val |= $_ << $bshift;
       }

       # 4 bytes (6 bits)
       $bshift = 24;
       $i = 0;
       for (1..4) {
          $bshift -= 6;
          $out .= $ENC[($val >> $bshift) & 0x3F];
       }
    }
    $out  = $ENC[$input_count] . "$out\n";
    return $out;
}

sub _decode {
    my $in = shift;
    my $out = '';

    my $total_octets = DEC_MAP(ord($in));
    my $end = ceil($total_octets/3) * 4;
    my $count_octets = 0;
    for (my $offset = 1; $offset < $end; $offset += 4) {
       my @chunk = unpack("C4", substr($in, $offset, 4));
       my ($bshift, $val, $i);

       # 4 bytes (6 bits)
       $bshift = 24;
       foreach (@chunk) {
          $bshift -= 6;
          $val |= DEC_MAP($_) << $bshift;
       }

       # 3 bytes (8 bits)
       $bshift = 24;
       $i = 0;
       while($i < 3 && $count_octets < $total_octets) {
          $bshift -= 8;
          $out .= chr(($val >> $bshift) & 0xFF);
          $count_octets ++;
          $i ++;
       }
    }
    return $out;
}

#------------------------------
#
# last_preamble
#
# Return the last preamble as ref to array of lines.
# Gets reset by decode_it().
#
sub last_preamble {
    my $self = shift;
    return $self->{MDX_Preamble} || [];
}

#------------------------------
#
# last_mode
#
# Return the last mode.
# Gets reset to undef by decode_it().
#
sub last_mode {
    shift->{MDX_Mode};
}

#------------------------------
#
# last_filename
#
# Return the last filename.
# Gets reset by decode_it().
#
sub last_filename {
    shift->{MDX_File} || [];
}

#------------------------------
1;
